@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2017-2018, ARM Limited and Contributors. All rights reserved.
+ * Copyright (c) 2017-2025, Arm Limited and Contributors. All rights reserved.
  *
  * SPDX-License-Identifier: BSD-3-Clause
  */
@@ -12,8 +12,10 @@
 
 #include <platform_def.h>
 
+#include <arch_features.h>
 #include <arch_helpers.h>
 #include <common/debug.h>
+#include <lib/utils.h>
 #include <lib/utils_def.h>
 #include <lib/xlat_tables/xlat_tables_defs.h>
 #include <lib/xlat_tables/xlat_tables_v2.h>
@@ -23,8 +25,9 @@
 /* Helper function that cleans the data cache only if it is enabled. */
 static inline __attribute__((unused)) void xlat_clean_dcache_range(uintptr_t addr, size_t size)
 {
-	if (is_dcache_enabled())
+	if (is_dcache_enabled()) {
 		clean_dcache_range(addr, size);
+	}
 }
 
 #if PLAT_XLAT_TABLES_DYNAMIC
@@ -108,6 +111,7 @@ uint64_t xlat_desc(const xlat_ctx_t *ctx, uint32_t attr,
 {
 	uint64_t desc;
 	uint32_t mem_type;
+	uint32_t shareability_type;
 
 	/* Make sure that the granularity is fine enough to map this address. */
 	assert((addr_pa & XLAT_BLOCK_MASK(level)) == 0U);
@@ -123,11 +127,14 @@ uint64_t xlat_desc(const xlat_ctx_t *ctx, uint32_t attr,
 	 * faults aren't managed.
 	 */
 	desc |= LOWER_ATTRS(ACCESS_FLAG);
+
+	/* Determine the physical address space this region belongs to. */
+	desc |= xlat_arch_get_pas(attr);
+
 	/*
-	 * Deduce other fields of the descriptor based on the MT_NS and MT_RW
-	 * memory region attributes.
+	 * Deduce other fields of the descriptor based on the MT_RW memory
+	 * region attributes.
 	 */
-	desc |= ((attr & MT_NS) != 0U) ? LOWER_ATTRS(NS) : 0U;
 	desc |= ((attr & MT_RW) != 0U) ? LOWER_ATTRS(AP_RW) : LOWER_ATTRS(AP_RO);
 
 	/*
@@ -193,8 +200,25 @@ uint64_t xlat_desc(const xlat_ctx_t *ctx, uint32_t attr,
 			desc |= xlat_arch_regime_get_xn_desc(ctx->xlat_regime);
 		}
 
+		shareability_type = MT_SHAREABILITY(attr);
 		if (mem_type == MT_MEMORY) {
-			desc |= LOWER_ATTRS(ATTR_IWBWA_OWBWA_NTR_INDEX | ISH);
+			desc |= LOWER_ATTRS(ATTR_IWBWA_OWBWA_NTR_INDEX);
+			if (shareability_type == MT_SHAREABILITY_NSH) {
+				desc |= LOWER_ATTRS(NSH);
+			} else if (shareability_type == MT_SHAREABILITY_OSH) {
+				desc |= LOWER_ATTRS(OSH);
+			} else {
+				desc |= LOWER_ATTRS(ISH);
+			}
+
+			/* Set GP bit for block and page code entries
+			 * if BTI mechanism is implemented.
+			 */
+			if (is_feat_bti_supported() &&
+			   ((attr & (MT_TYPE_MASK | MT_RW |
+				MT_EXECUTE_NEVER)) == MT_CODE)) {
+				desc |= GP;
+			}
 		} else {
 			assert(mem_type == MT_NON_CACHEABLE);
 			desc |= LOWER_ATTRS(ATTR_NON_CACHEABLE_INDEX | OSH);
@@ -472,10 +496,11 @@ static action_t xlat_tables_map_region_action(const mmap_region_t *mm,
 				 */
 				if (((dest_pa & XLAT_BLOCK_MASK(level)) != 0U)
 				    || (level < MIN_LVL_BLOCK_DESC) ||
-				    (mm->granularity < XLAT_BLOCK_SIZE(level)))
+				    (mm->granularity < XLAT_BLOCK_SIZE(level))) {
 					return ACTION_CREATE_NEW_TABLE;
-				else
+				} else {
 					return ACTION_WRITE_BLOCK_ENTRY;
+				}
 
 			} else {
 				/*
@@ -594,7 +619,8 @@ static uintptr_t xlat_tables_map_region(xlat_ctx_t *ctx, mmap_region_t *mm,
 			}
 
 			/* Point to new subtable from this one. */
-			table_base[table_idx] = TABLE_DESC | (unsigned long)subtable;
+			table_base[table_idx] =
+				TABLE_DESC | (uintptr_t)subtable;
 
 			/* Recurse to write into subtable */
 			end_va = xlat_tables_map_region(ctx, mm, table_idx_va,
@@ -605,8 +631,9 @@ static uintptr_t xlat_tables_map_region(xlat_ctx_t *ctx, mmap_region_t *mm,
 				XLAT_TABLE_ENTRIES * sizeof(uint64_t));
 #endif
 			if (end_va !=
-				(table_idx_va + XLAT_BLOCK_SIZE(level) - 1U))
+				(table_idx_va + XLAT_BLOCK_SIZE(level) - 1U)) {
 				return end_va;
+			}
 
 		} else if (action == ACTION_RECURSE_INTO_TABLE) {
 			uintptr_t end_va;
@@ -621,8 +648,9 @@ static uintptr_t xlat_tables_map_region(xlat_ctx_t *ctx, mmap_region_t *mm,
 				XLAT_TABLE_ENTRIES * sizeof(uint64_t));
 #endif
 			if (end_va !=
-				(table_idx_va + XLAT_BLOCK_SIZE(level) - 1U))
+				(table_idx_va + XLAT_BLOCK_SIZE(level) - 1U)) {
 				return end_va;
+			}
 
 		} else {
 
@@ -634,8 +662,9 @@ static uintptr_t xlat_tables_map_region(xlat_ctx_t *ctx, mmap_region_t *mm,
 		table_idx_va += XLAT_BLOCK_SIZE(level);
 
 		/* If reached the end of the region, exit */
-		if (mm_end_va <= table_idx_va)
+		if (mm_end_va <= table_idx_va) {
 			break;
+		}
 	}
 
 	return table_idx_va - 1U;
@@ -661,8 +690,9 @@ static int mmap_add_region_check(const xlat_ctx_t *ctx, const mmap_region_t *mm)
 	uintptr_t end_va = base_va + size - 1U;
 
 	if (!IS_PAGE_ALIGNED(base_pa) || !IS_PAGE_ALIGNED(base_va) ||
-			!IS_PAGE_ALIGNED(size))
+			!IS_PAGE_ALIGNED(size)) {
 		return -EINVAL;
+	}
 
 	if ((granularity != XLAT_BLOCK_SIZE(1U)) &&
 		(granularity != XLAT_BLOCK_SIZE(2U)) &&
@@ -671,18 +701,22 @@ static int mmap_add_region_check(const xlat_ctx_t *ctx, const mmap_region_t *mm)
 	}
 
 	/* Check for overflows */
-	if ((base_pa > end_pa) || (base_va > end_va))
+	if ((base_pa > end_pa) || (base_va > end_va)) {
 		return -ERANGE;
+	}
 
-	if ((base_va + (uintptr_t)size - (uintptr_t)1) > ctx->va_max_address)
+	if (end_va > ctx->va_max_address) {
 		return -ERANGE;
+	}
 
-	if ((base_pa + (unsigned long long)size - 1ULL) > ctx->pa_max_address)
+	if (end_pa > ctx->pa_max_address) {
 		return -ERANGE;
+	}
 
 	/* Check that there is space in the ctx->mmap array */
-	if (ctx->mmap[ctx->mmap_num - 1].size != 0U)
+	if (ctx->mmap[ctx->mmap_num - 1].size != 0U) {
 		return -ENOMEM;
+	}
 
 	/* Check for PAs and VAs overlaps with all other regions */
 	for (const mmap_region_t *mm_cursor = ctx->mmap;
@@ -711,16 +745,19 @@ static int mmap_add_region_check(const xlat_ctx_t *ctx, const mmap_region_t *mm)
 
 #if PLAT_XLAT_TABLES_DYNAMIC
 			if (((mm->attr & MT_DYNAMIC) != 0U) ||
-			    ((mm_cursor->attr & MT_DYNAMIC) != 0U))
+			    ((mm_cursor->attr & MT_DYNAMIC) != 0U)) {
 				return -EPERM;
+			}
 #endif /* PLAT_XLAT_TABLES_DYNAMIC */
 			if ((mm_cursor->base_va - mm_cursor->base_pa) !=
-							(base_va - base_pa))
+							(base_va - base_pa)) {
 				return -EPERM;
+			}
 
 			if ((base_va == mm_cursor->base_va) &&
-						(size == mm_cursor->size))
+						(size == mm_cursor->size)) {
 				return -EPERM;
+			}
 
 		} else {
 			/*
@@ -737,8 +774,9 @@ static int mmap_add_region_check(const xlat_ctx_t *ctx, const mmap_region_t *mm)
 			bool separated_va = (end_va < mm_cursor->base_va) ||
 				(base_va > mm_cursor_end_va);
 
-			if (!separated_va || !separated_pa)
+			if (!separated_va || !separated_pa) {
 				return -EPERM;
+			}
 		}
 	}
 
@@ -755,8 +793,9 @@ void mmap_add_region_ctx(xlat_ctx_t *ctx, const mmap_region_t *mm)
 	int ret;
 
 	/* Ignore empty regions */
-	if (mm->size == 0U)
+	if (mm->size == 0U) {
 		return;
+	}
 
 	/* Static regions must be added before initializing the xlat tables. */
 	assert(!ctx->initialized);
@@ -831,10 +870,13 @@ void mmap_add_region_ctx(xlat_ctx_t *ctx, const mmap_region_t *mm)
 
 	*mm_cursor = *mm;
 
-	if (end_pa > ctx->max_pa)
+	if (end_pa > ctx->max_pa) {
 		ctx->max_pa = end_pa;
-	if (end_va > ctx->max_va)
+	}
+
+	if (end_va > ctx->max_va) {
 		ctx->max_va = end_va;
+	}
 }
 
 /*
@@ -889,8 +931,9 @@ static void mmap_alloc_va_align_ctx(xlat_ctx_t *ctx, mmap_region_t *mm)
 	 */
 	for (unsigned int level = ctx->base_level; level <= 2U; ++level) {
 
-		if ((align_check & XLAT_BLOCK_MASK(level)) != 0U)
+		if ((align_check & XLAT_BLOCK_MASK(level)) != 0U) {
 			continue;
+		}
 
 		mm->base_va = round_up(mm->base_va, XLAT_BLOCK_SIZE(level));
 		return;
@@ -962,7 +1005,7 @@ int mmap_add_dynamic_region_ctx(xlat_ctx_t *ctx, mmap_region_t *mm)
 		     (uintptr_t)mm_last - (uintptr_t)mm_cursor);
 
 	/*
-	 * Check we haven't lost the empty sentinal from the end of the array.
+	 * Check we haven't lost the empty sentinel from the end of the array.
 	 * This shouldn't happen as we have checked in mmap_add_region_check
 	 * that there is free space.
 	 */
@@ -1178,16 +1221,13 @@ void __init init_xlat_tables_ctx(xlat_ctx_t *ctx)
 	xlat_mmap_print(mm);
 
 	/* All tables must be zeroed before mapping any region. */
+	zeromem(ctx->base_table, ctx->base_table_entries * sizeof(uint64_t));
 
-	for (unsigned int i = 0U; i < ctx->base_table_entries; i++)
-		ctx->base_table[i] = INVALID_DESC;
-
-	for (int j = 0; j < ctx->tables_num; j++) {
 #if PLAT_XLAT_TABLES_DYNAMIC
-		ctx->tables_mapped_regions[j] = 0;
+	zeromem(ctx->tables_mapped_regions, ctx->tables_num * sizeof(uint32_t));
 #endif
-		for (unsigned int i = 0U; i < XLAT_TABLE_ENTRIES; i++)
-			ctx->tables[j][i] = INVALID_DESC;
+	for (int i = 0; i < ctx->tables_num; i++) {
+		zeromem(ctx->tables[i], XLAT_TABLE_ENTRIES * sizeof(uint64_t));
 	}
 
 	while (mm->size != 0U) {
